
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.0.17'
version = '1.0.17'
full_version = '1.0.17'
git_revision = 'e54dfa6a56df9d37a2035bf667a8449fe4ea2125'
release = True

if not release:
    version = full_version
    short_version += ".dev"
